const Protobuf = require('pbf');
const VectorTile = require('vector-tile').VectorTile;
const gdal = require('gdal');
const wkx = require('wkx');

const mercatorProj = '+proj=merc +a=6378137 +b=6378137 +lat_ts=0.0 +lon_0=0.0 +x_0=0.0 +y_0=0.0 +k=1.0 +units=m +nadgrids=@null +wktext +no_defs +over';
const WGS84Proj = '+proj=longlat +ellps=WGS84 +datum=WGS84 +no_defs';

String.prototype.replaceAll = function(find, replace) {
    var str = this;
    return str.replace(new RegExp(find.replace(/([.*+?^=!:${}()|\[\]\/\\])/g, "\\$1"), 'g'), replace);
};
module.exports = {
    genAndClipThai: function(data, x, y, z, tilegeo, geoThai) {
        var tile = new VectorTile(new Protobuf(data)); 
        var clipThai = tilegeo.intersection(geoThai);
        var listLayers = [];
        for(var layerId in tile.layers){
            var layer = tile.layers[layerId];
            var isBoundary = layer.name == 'admin';

            var list = []; 
            for (var i = 0; i < layer.length; i++) {
                var feature = layer.feature(i);
            
                var json = feature.toGeoJSON(x,y,z);
                if(isBoundary&&json.properties['admin_level']<=2){
                    list.push(json);
                }else if(layer.name == 'water'){
                    var geo = wkx.Geometry.parseGeoJSON(json.geometry);
                    var wkb = geo.toWkb();
                    var geometry = gdal.Geometry.fromWKB(wkb);
                    if(!geoThai.contains(geometry)){
                        list.push(json);
                    }
                    //console.log(json.properties);
                }else{
                    var geo = wkx.Geometry.parseGeoJSON(json.geometry);
                    var wkb = geo.toWkb();
                    var geometry = gdal.Geometry.fromWKB(wkb);
                    if(!clipThai.contains(geometry)){
                        if(json.geometry.type == 'Point'){
                            list.push(json);
                        }else{
                            var clip = geometry.difference(clipThai);
                            if(clip){
                                var clipJson = JSON.parse(clip.toJSON());
                                //console.log(clipJson);
                                if(clipJson.type != 'GeometryCollection'){
                                    var newJson = Object.assign({},json,{geometry: clipJson});
                                    //console.log(newJson);
                                    list.push(newJson);
                                }
                            }
                        }
                    }
                }
            }
            var geojson = {
                "type": "FeatureCollection",
                "features":list
            };
            listLayers.push({
                name: layer.name,
                srs: WGS84Proj,
                data: JSON.stringify(geojson)
            });
        }
        return listLayers;
    },
    gen(data, x, y, z){
        var tile = new VectorTile(new Protobuf(data)); 
        var listLayers = [];
        for(var layerId in tile.layers){
            var layer = tile.layers[layerId];
            var list = []; 
            for (var i = 0; i < layer.length; i++) {
                var feature = layer.feature(i);
                var json = feature.toGeoJSON(x,y,z);
                list.push(json);
            }
            var geojson = {
                "type": "FeatureCollection",
                "features":list
            };
            listLayers.push({
                name: layer.name,
                srs: WGS84Proj,
                data: JSON.stringify(geojson)
            });
        }
        return listLayers;
    }
};
