#include <node.h>
#include <nan.h>
#include <iostream>
#include <fstream>
#include <string>
#include <vector>

#include "tnvectortilerender.hpp"
using namespace v8;
using namespace std;
//static TNVectorTileRender *tileRender = NULL;
static std::vector<TNVectorTileRender *> *listRender = NULL;
void renderVectorTile(const FunctionCallbackInfo<Value>& args) {
  Isolate* isolate = args.GetIsolate();
  if(args.Length() < 4){
    isolate->ThrowException(Exception::TypeError(
        String::NewFromUtf8(isolate, "Wrong number of arguments")));
    return;
  }
  int index = args[0]->NumberValue();
  int x = args[1]->NumberValue();
  int y = args[2]->NumberValue();
  int z = args[3]->NumberValue();
  
  
  Local<Function> cb = Local<Function>::Cast(args[4]);
  TNVectorTileRender *tileRender = listRender->at(index);
  if(tileRender){
      std::string buffer = tileRender->renderVectorTile(x,y,z,true);
      const unsigned argc = 1;
      const char* buf = buffer.c_str();
      size_t len = buffer.length();

      Local<Value> argv[argc] =  {Nan::CopyBuffer(buf, len).ToLocalChecked()}; //{ String::NewFromUtf8(isolate, buffer.c_str()) };
      cb->Call(Null(isolate), argc, argv);
  }
}

TNGeoJsonLayerRender unpackLayer(Isolate * isolate, Local<Object> layerObj){
   TNGeoJsonLayerRender layer;
   Handle<Value> nameValue = layerObj->Get(String::NewFromUtf8(isolate,"name"));
   Handle<Value> srsValue = layerObj->Get(String::NewFromUtf8(isolate,"srs"));
   Handle<Value> dataValue = layerObj->Get(String::NewFromUtf8(isolate,"data"));

   String::Utf8Value nameString(nameValue);
   String::Utf8Value srsString(srsValue);
   String::Utf8Value dataString(dataValue);
   layer.name = std::string(*nameString); //nameString.tostr();
   layer.srs = std::string(*srsString);  //srsString.c_str();
   layer.data = std::string(*dataString);

   return layer;
}
void renderVectorTileAddGeoData(const FunctionCallbackInfo<Value>& args){
  Isolate* isolate = args.GetIsolate();
  if(args.Length() < 5){
    isolate->ThrowException(Exception::TypeError(
        String::NewFromUtf8(isolate, "Wrong number of arguments")));
    return;
  }
  int index = args[0]->NumberValue();
  int x = args[1]->NumberValue();
  int y = args[2]->NumberValue();
  int z = args[3]->NumberValue();

  Local<Array> inputLayer = Local<Array>::Cast(args[4]);
  unsigned int numLayer = inputLayer->Length();
  
  std::vector<TNGeoJsonLayerRender> layers;
  for (unsigned int i = 0; i < numLayer; i++) {
      TNGeoJsonLayerRender layer = unpackLayer(isolate, Local<Object>::Cast(inputLayer->Get(i)));
      layers.push_back(layer);
  }
  
  Local<Function> cb = Local<Function>::Cast(args[5]);
  TNVectorTileRender *tileRender = listRender->at(index);
  if(tileRender){
      int size = tileRender->addLayerWithGeoJson(layers);

      std::string buffer = tileRender->renderVectorTile(x,y,z,true);
      const unsigned argc = 1;
      const char* buf = buffer.c_str();
      size_t len = buffer.length();

      Local<Value> argv[argc] =  {Nan::CopyBuffer(buf, len).ToLocalChecked()}; //{ String::NewFromUtf8(isolate, buffer.c_str()) };
      
      tileRender->removeLayerWithSizeFromLast(size);
      cb->Call(Null(isolate), argc, argv);
  }
}
void initRenderVectorTile(const FunctionCallbackInfo<Value>& args){
   if(listRender == NULL){
      listRender = new std::vector<TNVectorTileRender *>();
   }
  String::Utf8Value param1(args[0]->ToString());
  String::Utf8Value param2(args[1]->ToString());
  std::string mapnikDir = std::string(*param1);
  std::string fontsDir = std::string(*param2);

  const char *xml = NULL;
  if(args.Length()==3){
     String::Utf8Value param3(args[2]->ToString());
     std::string xmlPath = std::string(*param3);
     xml = xmlPath.c_str();
  }
  TNVectorTileRender *tileRender = new TNVectorTileRender(mapnikDir.c_str(),fontsDir.c_str(),xml);

  listRender->push_back(tileRender);
}
void helloVectorTile(const FunctionCallbackInfo<Value>& args){
  Isolate* isolate = args.GetIsolate();
  Local<Function> cb = Local<Function>::Cast(args[3]);
  const unsigned argc = 1;
  Local<Value> argv[argc] = { String::NewFromUtf8(isolate, "hello world") };
  cb->Call(Null(isolate), argc, argv);
}
void loadStyleVectorTile(const FunctionCallbackInfo<Value>& args){
  int index = args[0]->NumberValue();
  TNVectorTileRender *tileRender = listRender->at(index);
  if(tileRender){
     String::Utf8Value param1(args[1]->ToString());
     std::string style = std::string(*param1);
     tileRender->loadMapWithString(style.c_str());
  }
}
void init(Local<Object> exports) {
  NODE_SET_METHOD(exports, "init", initRenderVectorTile);
  NODE_SET_METHOD(exports, "loadStyle", loadStyleVectorTile);
  NODE_SET_METHOD(exports, "render", renderVectorTile);
  NODE_SET_METHOD(exports, "renderWithGeoData", renderVectorTileAddGeoData);
  NODE_SET_METHOD(exports, "hello",helloVectorTile);

}

NODE_MODULE(addon, init)
