/*
 *  tnvectortilerender.cpp
 *  tnvectortilerender
 *
 *  Created by isaret on 10/14/16.
 *  Copyright © 2016 THiNKNET. All rights reserved.
 *
 */

#include <iostream>
#include "tnvectortilerender.hpp"

#include <iostream>
#include <fstream>
#include <string>
#include <vector>

// mapnik
#include <mapnik/map.hpp>
#include <mapnik/load_map.hpp>
#include <mapnik/value_error.hpp>
#include <mapnik/datasource_cache.hpp>
#include <mapnik/box2d.hpp>
#include <mapnik/proj_transform.hpp>
#include <mapnik/image_util.hpp>

#include <mapnik/layer.hpp>
#include <mapnik/datasource_cache.hpp>

// mapnik vector tiles generation
#include <vector_tile.pb.h>
#include <vector_tile_projection.hpp>
#include <vector_tile_processor.hpp>
#include <vector_tile_compression.hpp>
//#include <vector_tile_backend_pbf.hpp>
using namespace std;
using namespace mapnik;

static int tile_size = 256;

class TNVectorTileRenderImpl{
public:
    TNVectorTileRenderImpl(const char *mapnikPath,const char *fontsPath ,const char *xmlPath);
    ~TNVectorTileRenderImpl();
    
    string renderVectorTile(int x,int y,int z,bool compressed);
    void loadMapWithString(const char * str);
    
    int addLayerWithGeoJson(std::vector<TNGeoJsonLayerRender> list);
    void removeLayerWithSizeFromLast(int size);
private:
    
private:
    Map titleMap;
};

TNVectorTileRender::TNVectorTileRender(const char *mapnikPath, const char *fontsPath, const char *xmlPath){
    this->render = new TNVectorTileRenderImpl(mapnikPath, fontsPath, xmlPath);
}
TNVectorTileRender::~TNVectorTileRender(){
    delete render;
}
string TNVectorTileRender::renderVectorTile(int x,int y,int z,bool compressed){
    return this->render->renderVectorTile(x, y, z, compressed);
}

void TNVectorTileRender::loadMapWithString(const char *str){
    this->render->loadMapWithString(str);
}
int TNVectorTileRender::addLayerWithGeoJson(std::vector<TNGeoJsonLayerRender> list){
    return this->render->addLayerWithGeoJson(list);
}
void TNVectorTileRender::removeLayerWithSizeFromLast(int size){
    this->render->removeLayerWithSizeFromLast(size);
}

TNVectorTileRenderImpl::TNVectorTileRenderImpl(const char *mapnikPath,const char *fontsPath,const char *xmlPath){
    datasource_cache::instance().register_datasources(mapnikPath);
    mapnik::freetype_engine::register_fonts(fontsPath);
    GOOGLE_PROTOBUF_VERIFY_VERSION;
    
    if(xmlPath){
        Map map(tile_size, tile_size, MAPNIK_GMERC_PROJ);
        load_map(map, xmlPath, true);
        this->titleMap = map;
    }
}
TNVectorTileRenderImpl::~TNVectorTileRenderImpl(){
    google::protobuf::ShutdownProtobufLibrary();
}
void TNVectorTileRenderImpl::loadMapWithString(const char * str){
    Map map(tile_size, tile_size, MAPNIK_GMERC_PROJ);
    load_map_string(map, str, true);
    this->titleMap = map;
}

int TNVectorTileRenderImpl::addLayerWithGeoJson(std::vector<TNGeoJsonLayerRender> list){
    int size = (int)list.size();
    
    for(int i=0;i<size;i++){
        TNGeoJsonLayerRender layer = list.at(i);
        
        parameters params;
        mapnik::layer lyr(layer.name, layer.srs);
        
        //cout << layer.data.c_str()<<"\n";
        
        params["type"] = "geojson";
        params["inline"] = layer.data.c_str();
        
        std::shared_ptr<datasource> ds = datasource_cache::instance().create(params);
        lyr.set_datasource(ds);
        this->titleMap.add_layer(lyr);
    }
    return size;
}
void TNVectorTileRenderImpl::removeLayerWithSizeFromLast(int size){
    int lastIndex = (int)this->titleMap.layer_count() -1;
    for(int i = 0;i<size; i++){
        this->titleMap.remove_layer(lastIndex-i);
    }
}
string TNVectorTileRenderImpl::renderVectorTile(int x,int y,int z,bool compression){
    
    mapnik::vector_tile_impl::processor ren(titleMap);
    //mapnik::vector_tile_impl::tile out_tile = ren.create_tile(x,y,z,40960,tile_size*2);
    mapnik::vector_tile_impl::tile out_tile = ren.create_tile(x,y,z,40960,tile_size*2);
    
    vector_tile::Tile tile;
    tile.ParseFromString(out_tile.get_buffer());
    
    string buffer,compressed;
    tile.SerializeToString(&buffer);
    if (compression) {
        mapnik::vector_tile_impl::zlib_compress(buffer, compressed);
        buffer = compressed;
        cout << ">>Creating compressed tile for z = " << z << " ; x = " << x << " ; y = " << y << "\n" ;
    } else {
        cout << ">>Creating tile for z = " << z << " ; x = " << x << " ; y = " << y << "\n";
    }
    //ofstream output("/Users/isaret/Documents/map/"+to_string(z) + "_" + to_string(x) + "_" + to_string(y) + "_lib.pbf");
    //output << buffer ;
    return buffer;
    
}
