const vectorTile = require('./build/Release/tnvectortile');
const zlib = require('zlib');
const fs = require('fs');
const gdal = require('gdal');
const wkx = require('wkx');
const s3 = require('s3');
const aws = require('aws-sdk');
const crc = require('crc');

const SphericalMercator = require('sphericalmercator');
const genMapLayer = require('./genmaplayer');
const mqRendering = require('./tnmqrendering');

const bufferSize = 256;
const maxZoomOSM = 14;
const ds = gdal.open('thailand_polygon.geojson');
const layerThailand = ds.layers.get(0);
const geoThailand = layerThailand.features.first().getGeometry();
const geoThailandOptimize = {};
const merc = new SphericalMercator({ size: bufferSize });
const maxTimeForRequestOnS3 = 5;
const timeForCheckTileOnS3 = 1000; //milisec
/*
config{
    awsConfig,
    renderConfig,
    awsClient,
    vectorTile,
    sourceOSM
}
mqdata{
    channel,
    msg
}
*/
const long2tile = function(lon,zoom) { return (Math.floor((lon+180)/360*Math.pow(2,zoom))); }
const lat2tile = function(lat,zoom)  { return (Math.floor((1-Math.log(Math.tan(lat*Math.PI/180) + 1/Math.cos(lat*Math.PI/180))/Math.PI)/2 *Math.pow(2,zoom))); }
module.exports = function(config, mqdata) {
    const findFeatureInGeoJsonWithId = function(featureId,features){
            for (var i=0; i < features.length; i++) {
                if (features[i].id === featureId) {
                    return i;
                }
            }
            return -1;
        }
        const convertGeoJsonToGeometry = function(json){
            var geo = wkx.Geometry.parseGeoJSON(json.geometry);
            var wkb = geo.toWkb();
            return gdal.Geometry.fromWKB(wkb);
        }
    const saveTileOnS3 = function(awsClient, awsParams, buffer, rabbitmqBucketName){
        awsParams['ContentType']   = 'application/x-protobuf';
        awsParams['Body']          = buffer;
        awsParams['ACL']           = 'private';
        awsClient.putObject(awsParams, function(err, data){
            if (err) {
                console.log(err, err.stack);
            }
            mqRendering.deleteTileRendering(awsClient, rabbitmqBucketName, awsParams['Key'], function(err, status){
                
            });
        });
    }
    const sendResponseToClient = function(awsClient, mqdata, tile, awsParams, rabbitmqBucketName){
        if(awsParams){
            saveTileOnS3(awsClient, awsParams, tile, rabbitmqBucketName);
        }
        var msg = mqdata.msg;
        var channel = mqdata.channel;
        var callback = mqdata.callback;
        if(callback){
            callback(null, tile);
        }else{
            channel.sendToQueue(msg.properties.replyTo,
                            tile,
                            {correlationId: msg.properties.correlationId, type:'pbf' });
            channel.ack(msg);
        }
    }
    const mergeMultiTile = function(config, mqdata, layers, x, y, z, awsParams){
        var maplayers = {};
        var sumLayers = [];
        for(var i =0;i<layers.length;i++){
            var layer = layers[i];
            var sum = maplayers[layers[i].name];
            if(sum){
                var index = sum.index;
                var jsonSumLayer = JSON.parse(sumLayers[index].data);
                var jsonLayer = JSON.parse(layer.data);
                    jsonLayer.features.map((feature,index) =>{
                        var indexFound = findFeatureInGeoJsonWithId(feature.id, jsonSumLayer.features);
                        if(indexFound!=-1){
                            var geo = feature.geometry;
                            if(geo.type!='Point'){
                                var geo = convertGeoJsonToGeometry(feature);
                                var sumGeo = convertGeoJsonToGeometry(jsonSumLayer.features[indexFound]);
                                var totalGeo = sumGeo.union(geo);
                                jsonSumLayer.features[indexFound].geometry = JSON.parse(totalGeo.toJSON());
                            }
                        }else{
                            //console.log('not have');
                            jsonSumLayer.features.push(feature);
                        }
                    });
                sumLayers[index] = Object.assign({},sumLayers[index],{data: JSON.stringify(jsonSumLayer)});
                //console.log(sumLayers[index].data);

            }else{
                maplayers[layers[i].name] = {index: sumLayers.length };
                sumLayers.push(layer);
            }
        }

        config.vectorTile.renderWithGeoData(1, x, y, z, sumLayers, function(tile){
            sendResponseToClient(config.awsClient, mqdata, tile, awsParams, config.awsConfig.rabbitmqBucketName);
        });
    }
    const loadMutiTile = function(config, mqdata, tilegeo, geoThai, list, layers, x, y, z, awsParams){
        var tile = list.pop();
         config.sourceOSM.getTile(maxZoomOSM, tile.x, tile.y, function(err, data, headers) {
            zlib.gunzip(data, function(err, buffer) {
                var layer = genMapLayer.genAndClipThai(buffer, tile.x, tile.y, maxZoomOSM, tilegeo, geoThai);
                layers = layers.concat(layer);

                if(list.length == 0){
                    mergeMultiTile(config, mqdata, layers, x, y, z, awsParams);
                }else{
                    loadMutiTile(config, mqdata, tilegeo, geoThai, list, layers, x, y, z, awsParams);
                }
            });
        });
    }
    const mergeOSMTile = function(config, mqdata, list, layers, x, y, z, awsParams){
        var tile = list.pop();
        config.sourceOSM.getTile(maxZoomOSM, tile.x, tile.y, function(err, data, headers) {
            zlib.gunzip(data, function(err, buffer) {
                var layer = genMapLayer.gen(buffer, tile.x, tile.y, maxZoomOSM);
                layers = layers.concat(layer);

                if(list.length == 0){
                    mergeMultiTile(config, mqdata, layers, x, y, z, awsParams);
                }else{
                    mergeOSMTile(config, mqdata, list, layers, x, y, z, awsParams);
                }
            });
        });
    }
    const getListTileMoreThanMaxZoomOSM = function(minx, miny, maxx, maxy){
        var minTileX = long2tile(minx, maxZoomOSM);
        var minTileY = lat2tile(maxy, maxZoomOSM);
        var maxTileX = long2tile(maxx, maxZoomOSM);
        var maxTileY = lat2tile(miny, maxZoomOSM);
        var listTiles = [];
        for(var ty = minTileY; ty <= maxTileY ; ty++){
            for(var tx = minTileX; tx <= maxTileX ;tx++){
                listTiles.push({x:tx, y:ty});
            }
        }
        return listTiles;
    }
    const renderTile = function(config, mqdata, x, y, z, awsParams){
        var box = merc.bbox(x,y,z);
        var minx = box[0];
        var miny = box[1];
        var maxx = box[2];
        var maxy = box[3];
        var wkt = 'POLYGON(('+minx+' '+maxy+','+maxx+' '+maxy+','+maxx+' '+miny+','+minx+' '+miny+','+minx+' '+maxy+'))';
        var tilegeo = gdal.Geometry.fromWKT(wkt);
        var geoThai = geoThailandOptimize[z.toString()];
        if(geoThai == null){
            var tolerance =  (maxx-minx)/bufferSize;
            geoThai = geoThailand.simplify(tolerance)
            geoThailandOptimize[z.toString()] = geoThai;
            console.log('zoom '+z+' = '+tolerance);
        }
        var status = 0;
        if(geoThai.contains(tilegeo)){
            status = 1;
        }else if(geoThai.intersects(tilegeo)){
            status = 2;
        }
        console.log('get vt tile %d, %d, %d, %s', z, x, y,status);

        if(status == 1){
            config.vectorTile.render(0, x, y, z, function(tile){
                sendResponseToClient(config.awsClient, mqdata, tile, awsParams, config.awsConfig.rabbitmqBucketName);
            });
        }else if(status == 2){
            config.sourceOSM.getTile(z, x, y, function(err, data, headers) {
                //console.log(data.getBody());
                if(err){
                    var listTiles = getListTileMoreThanMaxZoomOSM(minx, miny, maxx, maxy);
                    
                    var layers = [];
                    loadMutiTile(config, mqdata, tilegeo, geoThai, listTiles, layers, x, y, z, awsParams);
                }else{
                    zlib.gunzip(data, function(err, buffer) {
                        var layers = genMapLayer.genAndClipThai(buffer, x, y, z, tilegeo, geoThai);
                        config.vectorTile.renderWithGeoData(1, x, y, z, layers, function(tile){
                            sendResponseToClient(config.awsClient, mqdata, tile, awsParams, config.awsConfig.rabbitmqBucketName);
                        });
                    });
                }

            });
        }else{
            config.sourceOSM.getTile(z, x, y, function(err, data, headers) {
                //console.log(data.getBody());
                if(err){
                    //console.log('OSM: '+err+':'+z+','+x+','+y+','+status);
                    //sendResponseToClient(config.awsClient, mqdata, new Buffer(''), null);
                    var listTiles = getListTileMoreThanMaxZoomOSM(minx, miny, maxx, maxy);
                    
                    var layers = [];
                    mergeOSMTile(config, mqdata, listTiles, layers, x, y, z, awsParams);
                }else{
                    sendResponseToClient(config.awsClient, mqdata, data, awsParams, config.awsConfig.rabbitmqBucketName);
                }
            });
        }
    }
    const getTileOnS3AndWaiting = function(awsClient, awsParams, mqdata, count){
        awsClient.getObject(awsParams, function(err, awsdata) {
            if(err){
                count++;
                if(count<maxTimeForRequestOnS3){
                    console.log('Waiting(%s)...', awsParams['Key']);
                   setTimeout(getTileOnS3AndWaiting, timeForCheckTileOnS3, awsClient, awsParams, mqdata, count);
                }else{
                    console.log('Timeout Waiting(%s)...', awsParams['Key']);
                    sendResponseToClient(awsClient, mqdata, new Buffer(''), null, null);
                }
            }else{
                console.log('Is coming S3 (%s)...', awsParams['Key']);
                sendResponseToClient(awsClient, mqdata, awsdata['Body'], null, null);
            }
        });
    }
    const getTile = function(config, mqdata){
        var x = mqdata.requestdata.x;
        var y = mqdata.requestdata.y;
        var z = mqdata.requestdata.z;

        var key = z+'_'+x+'_'+y;
        var hashKey = crc.crc32(key);
        var keyPath = hashKey+'/'+ z + '/' + x + '/' + y + '.pbf';
        var awsParams = {
                        Bucket: config.awsConfig.awsBucketNameFormat.replace('{name}',config.renderConfig.name),
                        Key: keyPath
                    };
        config.awsClient.getObject(awsParams, function(err, awsdata) {
            if(err){
                var rabbitmqBucketName = config.awsConfig.rabbitmqBucketName;
                mqRendering.checkIsRendering(awsClient, rabbitmqBucketName, keyPath, function(err, isRendering){
                    if(isRendering){
                        var count = 0;
                        console.log('Is rendering vector tile');
                        setTimeout(getTileOnS3AndWaiting, timeForCheckTileOnS3, config.awsClient, awsParams, mqdata, count);
                    }else{
                        renderTile(config, mqdata, x, y, z, awsParams);
                    }
                });
            }else{
                console.log('vector tile have in s3');
                //res.send(awsdata['Body']);
                sendResponseToClient(config.awsClient, mqdata, awsdata['Body'], null, null);
            }
        });
    }
    return getTile(config, mqdata);
}
