const amqp = require('amqplib');
const fs = require('fs')
const s3 = require('s3');
const aws = require('aws-sdk');
const vectorTile = require('./build/Release/tnvectortile');
const renderVectorTile = require('./tnvectortile');
const renderRasterTile = require('./tnrastertile');
const config = require('./config.json');
module.exports = function(connect, osmSource) {
    const initVectorTile = function(){
        const vtRenderConfig = config.vector.styles[0];
        const vtAwsConfig = config.vector.aws;
        const style = fs.readFileSync(vtRenderConfig.style, "utf-8");

        vectorTile.init("/usr/lib/mapnik/3.0/input","/usr/lib/mapnik/3.0/input/fonts/");
        vectorTile.init("/usr/lib/mapnik/3.0/input","/usr/lib/mapnik/3.0/input/fonts/");
        vectorTile.loadStyle(0,style);
        vectorTile.loadStyle(1,style);
        const awsClient = new aws.S3({  accessKeyId: vtAwsConfig.awsAccessKeyId,
                                        secretAccessKey: vtAwsConfig.awsSecretAccessKey,
                                        region: vtAwsConfig.awsRegion});
         return {   awsConfig: vtAwsConfig,
                    renderConfig: vtRenderConfig,
                    awsClient: awsClient,
                    vectorTile: vectorTile,
                    sourceOSM: osmSource
                };
    }
    const initRasterTile = function(vtconfig){
        var renders = []
        config.raster.styles.map((style,index) => {
            renders[style.name] = {
                config: {
                    awsConfig: config.raster.aws,
                    renderConfig: style
                },
                render: renderRasterTile.initRasterRender(style, config.raster.aws, vtconfig)
            }
        });
        return renders;
    }

    const vtconfig = initVectorTile();
    const rtconfig = initRasterTile(vtconfig);

    //return renderTile(vectorTile, awsClient, osmSource, channel, msg);
    return connect.createChannel().then(function(ch) {
        var q = 'rpc_queue';
        var ok = ch.assertQueue(q, {durable: false});
        var ok = ok.then(function() {
        ch.prefetch(1);
            return ch.consume(q, reply);
        });
            return ok.then(function() {
            console.log(' [x] Awaiting RPC requests');
        });

        function reply(msg) {
            var requestdata = JSON.parse(msg.content.toString());
            var mqdata = {
                channel: ch,
                msg: msg,
                requestdata: requestdata
            }
            if(requestdata.type == 'pbf'){
                renderVectorTile(vtconfig, mqdata);
            }else if(requestdata.type == 'png'){
                var render = rtconfig[requestdata.style];
                renderRasterTile.render(render.config, mqdata, render.render);
            }
        }
    });
}